/* $Id: screen.c,v 1.4 2006/01/13 20:57:47 cegger Exp $
***************************************************************************

   X server for LibGGI - screen implementation

   Copyright (C) 1997	   Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-1999 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#define PSZ 8
#include <ggi/gg.h>

#include "mi.h"
#include "cfb.h"
#include "cfb16.h"
#include "cfb24.h"
#include "cfb32.h"

#include "scrnintstr.h"
#include "mibstore.h"
#include "gcstruct.h"

#include "xggi.h"
#include "xggi_misc.h"


static unsigned long mask2offset(unsigned long mask)
{
	int i;
	
	for (i = 0; (mask & 1) == 0; i++) {
		mask >>= 1;
	}	/* for */
	
	return i;
}	/* mask2offset */


static int xggiScreenReset(int scr)
{
	SCREEN(scr).ptrbuttons = DEFAULT_PTRBUTTONS;
	SCREEN(scr).numvisuals = 0;
	SCREEN(scr).visuals = NULL;

	SCREEN_SCR(scr) = NULL;
	SCREEN_BACKBUFFER(scr) = NULL;
	SCREEN_ISMAPPED(scr) = 1;

	SCREEN_SCRIDX(scr) = -1;
	SCREEN_GCIDX(scr) = -1;

	return 0;
}	/* xggiScreenReset */


int xggiScreenAlloc(int numvisuals)
{
	int i;

	xggiScreens = safe_alloc(sizeof(struct xggiScreenInfo)
				* xggiInfo.numscreens);

	for (i = 0; i < xggiInfo.numscreens; i++) {
		xggiScreenReset(i);
		xggiVisualAlloc(&SCREEN(i), numvisuals);
	}	/* for */

	return 0;
}	/* xggiScreenAlloc */


int xggiScreenFree(void)
{
	int i;

	if (xggiScreens == NULL) {
		ErrorF("BUG: xggiScreenFree() called multiple times!\n");
	}	/* if */

	for (i = 0; i < xggiInfo.numscreens; i++) {
		xggiVisualFree(&(SCREEN(i)));
	}	/* for */

	xfree(xggiScreens);
	xggiScreens = NULL;
	xggiInfo.numscreens = xggiInfo.rows = xggiInfo.columns = 0;

	return 0;
}	/* xggiScreenFree */



static Bool xggiAllocatePrivates(int scr)
{

	SCREEN_SCRIDX(scr) = AllocateScreenPrivateIndex();
	if (SCREEN_SCRIDX(scr) < 0) {
		ErrorF("xggiAllocatePrivates: AllocateScreenPrivateIndex() failed\n");
		return FALSE;
	}

	SCREEN_GCIDX(scr) = AllocateGCPrivateIndex();
	if (SCREEN_GCIDX(scr) < 0) {
		ErrorF("xggiAllocatePrivates: AllocateGCPrivateIndex() failed\n");
		return FALSE;
	}	/* if */

	if (!AllocateGCPrivate(SCREEN_SCR(scr), SCREEN_GCIDX(scr), sizeof(xggiGCPrivRec))) {
		ErrorF("xggiAllocatePrivates: AllocateGCPrivate() failed\n");
		return FALSE;
	}	/* if */

	/* SCREEN_SCRBACKUP(scr) */
	SCREEN_SCRPRIV(scr).ptr = (pointer)xalloc(sizeof(xggiScreenBackupRec));
	if (!SCREEN_SCRBACKUP(scr)) {
		ErrorF("xggiScreenReset: screen private allocation failed\n");
		return FALSE;
	}	/* if */

	return TRUE;
}	/* xggiAllocatePrivates */



/*
 * xggiScreenInit
 *
 */
Bool xggiScreenInit(ScreenPtr pScreen, pointer pbits,
		int xsize, int ysize, int dpix, int dpiy, int width)
{
	VisualPtr visuals;
	DepthPtr depths;
	int nvisuals;
	int ndepths;
	int rootdepth = 0;
	VisualID defaultVisual;
	int bitsPerRGB;
	int	i;
	Bool Rstatus;
	VisualPtr visual;
	int scr;

	scr = xggiGetScreenIdx(pScreen);

#ifdef XGGI_DEBUG
	ErrorF("scr: %d, depth: %d, bpp:%d\n", scr,
		MASTER_VISUAL_DEPTH(scr),
		MASTER_VISUAL_BPP(scr));
#endif

	switch (MASTER_VISUAL_DEPTH(scr)) {
	case 8:
		bitsPerRGB = 8;
		break;
	case 15:
		bitsPerRGB = 5;
		break;
	case 16:
		bitsPerRGB = 6;
		break;
	default:
		bitsPerRGB = 8;
		break;
	}

	if (MASTER_VISUAL_BPP(scr) > 8) {
		if (!cfbSetVisualTypes(MASTER_VISUAL_DEPTH(scr),
				       1 << TrueColor, bitsPerRGB))
		{
			ErrorF("cfbSetVisualTypes failed\n");
			return FALSE;
		}
	}
	if (!cfbInitVisuals(&visuals, &depths, &nvisuals, 
			    &ndepths, &rootdepth, &defaultVisual,
			    1U << (MASTER_VISUAL_BPP(scr) - 1),
			    bitsPerRGB))
	{
		ErrorF("cfbInitVisuals failed\n");
		return FALSE;
	}

	SCREEN_ROOTDEPTH(scr) = rootdepth;

#ifdef XGGI_DEBUG
	ErrorF("rootdepth: %d\n", SCREEN_ROOTDEPTH(scr));
#endif

	if (SCREEN_ROOTDEPTH(scr) > 8) {
		for (i = 0, visual = visuals; i < nvisuals; i++, visual++) {
			if (visual->class == DirectColor
			    || visual->class == TrueColor) {
#define PIXFMT(scr)	(MASTER_VISUAL(scr).dbuf->buffer.plb.pixelformat)
				visual->redMask = PIXFMT(scr)->red_mask;
				visual->greenMask = PIXFMT(scr)->green_mask;
				visual->blueMask = PIXFMT(scr)->blue_mask;
#undef PIXFMT
				visual->offsetRed
					= mask2offset(visual->redMask);
				visual->offsetGreen
					= mask2offset(visual->greenMask);
				visual->offsetBlue
					= mask2offset(visual->blueMask);
			}
		}
	}

	if (scr == 0) {
		cfbWindowPrivateIndex = cfbGCPrivateIndex = -1;
	}
	pScreen->defColormap = FakeClientID(0);
	pScreen->whitePixel = (Pixel) 1;
	pScreen->blackPixel = (Pixel) 0;
	pScreen->QueryBestSize = mfbQueryBestSize;
	pScreen->RealizeFont = mfbRealizeFont;
	pScreen->UnrealizeFont = mfbUnrealizeFont;


	Rstatus = miScreenInit(pScreen, pbits, xsize, ysize, dpix, dpiy, width,
			       SCREEN_ROOTDEPTH(scr), ndepths, depths, defaultVisual,
			       nvisuals, visuals);

	if (xggiInfo.XGGIGeneration != serverGeneration) {
		xggiAllocatePrivates(scr);
		xggiAccelInit(scr);

		xggiInfo.XGGIGeneration = serverGeneration;
	}

	return Rstatus;
}	/* xggiScreenInit */




int xggiGetScreenIdx(ScreenPtr screen)
{
	int scr;

	for (scr = 0; scr < xggiInfo.numscreens; scr++) {
		if (SCREEN_SCR(scr) == screen) return scr;
	}

#if 0	/* What's this? */
	*((uint32_t*)0) = 43;
#endif
	FatalError("Tried to operate on invalid screen\n");

	/* Never reached */
	return -1;
}	/* xggiGetScreenIdx */

